# -*- coding: latin1 -*-
#
# Copyright (C) 2003 Ayose Cazorla Len
#
# Authors
#       Ayose Cazorla <ayose.cazorla@hispalinux.es>
#
# This file is part of Pyslide.
#
# Pyslide is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Pyslide is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyslide; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import pygame

from pygame.locals import *

class Effect(pygame.sprite.Sprite):
    '''Effect(for_open, parent, attrs)

    This is the base class for all the effects
    '''

    def __init__(self, for_open, parent, attrs):
        pygame.sprite.Sprite.__init__(self)
        self.raw_attrs = attrs
        self.parent = parent
        self.for_open = for_open
        self.parse_attrs(attrs)

    def parse_attrs(self, attrs):
        '''parse_attrs(attrs)

        Load the effect attributes
        '''
        pass

    def start(self):
        '''start()

        Starts the effects. The effect will modify
        the image and rect attributes of the parent
        '''

        import time
        self.start_time = time.time()

    def stop(self):
        '''stop()

        Restores the image and rect of the parent and
        remove itself from all the groups
        '''
        self.kill()

    def update(self):
        '''update() -> None

        The sprite.update event
        '''
        pass

class MoveEffect(Effect):
    '''MoveEffect

    Implements the Move effect
    '''
    def parse_attrs(self, attrs):
        'See Effect.parse_attrs for info'

        import misc, Presentation
        try:
            x, y, t = [int(x) for x in attrs.split(',')]
        except ValueError:
            self.dx, self.dy, self.time = 0

            raise misc.PyslideError, 'Incorrect move arguments'
        else:
            self.dx, self.dy = Presentation.CurrentSurface.map_units(x, y)
            self.time = t / 10.

    def start(self):
        'See Effect.start for info.'

        Effect.start(self)
        self.original_rect = self.parent.rect
        if self.for_open:
            self.parent.rect = self.parent.rect.move(-self.dx, -self.dy)

    def stop(self):
        Effect.stop(self)
        self.parent.rect = self.original_rect
        del self.original_rect

    def update(self, *k):
        import time
        delta = time.time() - self.start_time

        if delta < self.time:
            ox, oy = self.parent.orig_x, self.parent.orig_y
            if not self.for_open:
                ox += self.dx
                oy += self.dy
            cx = ox - self.dx * (self.time - delta) / self.time
            cy = oy - self.dy * (self.time - delta) / self.time
            self.parent.rect = pygame.Rect((cx, cy) + self.original_rect.size)
        else:
            self.stop()

class AlphaEffect(Effect):

    def parse_attrs(self, attrs):
        try:
            n = int(attrs)
        except ValueError:
            self.time = 1

            import misc
            raise misc.PyslideError, 'Invalid value for Alpha effect'
        else:
            self.time = n / 10.

    def start(self):
        Effect.start(self)

        # start with transparent or opaque
        if self.for_open:
            self.last_value = 0
        else:
            self.last_value = 255

        # Create a new surface only if we need to change
        # its alpha pixles

        if (self.parent.image.get_masks()[3] != 0):
            self.alpha_pixels = True

            image = pygame.Surface(self.parent.image.get_size())
            if image.get_masks()[3] == 0:
                image = image.convert_alpha() # we need the alpha
            image.fill(0)

            if self.last_value != 0:
                image.blit(self.parent.image, (0,0))

            self.original_image = self.parent.image
            self.parent.image = image

            # the same colorkey
            ck = self.original_image.get_colorkey()
            if ck is not None:
                image.set_colorkey(ck)

        else:
            self.alpha_pixels = False
            self.original_alpha_value = self.parent.image.get_alpha()
            self.parent.image.set_alpha(self.last_value)

    def stop(self):
        Effect.stop(self)
        # restore only is this effect is for_open. This is because
        # if the effect is for close, we'd to keep the image hidden
        if self.for_open:
            if self.alpha_pixels:
                self.parent.image = self.original_image
            elif self.original_alpha_value is not None:
                self.parent.image.set_alpha(self.original_alpha_value)
            else:
                self.parent.image.set_alpha(255)

    def update(self, *k):
        del k
        import time

        delta = (time.time() - self.start_time) / self.time

        if delta >= 1.:
            self.stop()
        else:
            self.apply_alpha(delta)

    def apply_alpha(self, delta):
        alpha = delta
        if not self.for_open: alpha = 1 - alpha

        if abs(alpha - self.last_value) > 0.01:
            self.last_value = alpha
            if self.alpha_pixels:
                import misc
                self.parent.image.blit(self.original_image, (0,0))
                misc.set_alpha(self.parent.image, alpha)
            else:
                self.parent.image.set_alpha(alpha*255)


class HVAlphaEffect(AlphaEffect):
    def __init__(self, *k):
        AlphaEffect.__init__(self, *k)
        self.is_vertical = None

    def parse_attrs(self, attrs):
        AlphaEffect.parse_attrs(self, attrs)
        if self.time < 0:
            self.direction = -1
            self.time = -self.time
        else:
            self.direction = 1

    def start(self):
        Effect.start(self)

        pi = self.parent.image
        self.alpha_pixels = True
        image = pygame.surface.Surface(pi.get_size()).convert_alpha()

        image.fill(0)
        if self.for_open:
            self.last_value = 0
        else:
            self.last_value = 255
            image.blit(pi, (0,0))

        self.parent.image = image
        self.original_image = pi

        self.limit = self.parent.image.get_size() \
                     [int(self.is_vertical)] + 100

    def apply_alpha(self, delta):
        import misc

        # Adjust the limit acording to the effect.
        d = int(delta * self.limit)

        # IF for_open XNOR direction < 0
        if self.for_open == (self.direction < 0):
            d = self.limit - d - 100

        # Change the alpha
        self.parent.image.fill(0)
        self.parent.image.blit(self.original_image, (0,0))
        if self.is_vertical:
            misc.set_valpha(self.parent.image, d, self.direction)
        else:
            misc.set_halpha(self.parent.image, d, self.direction)

class VAlphaEffect(HVAlphaEffect):
    def __init__(self, *k):
        HVAlphaEffect.__init__(self, *k)
        self.is_vertical = True

class HAlphaEffect(HVAlphaEffect):
    def __init__(self, *k):
        HVAlphaEffect.__init__(self, *k)
        self.is_vertical = False

class EffectError(Exception):
    pass

def load(for_open, parent, name, params):
    '''load(parent, for_open, name, params) -> Effect

    Loads an effect. If the effect is not available, the EffectError
    exception is raised.
    '''

    try:
        e = available_effects[name]
    except KeyError:
        raise EffectError, 'There is no "%s" effect' % name

    return e(for_open, parent, params)



available_effects = {
    'move': MoveEffect,
    'alpha': AlphaEffect,
    'valpha': VAlphaEffect,
    'halpha': HAlphaEffect,
};

