# -*- coding: latin1 -*-
#
# Copyright (C) 2003 Ayose Cazorla Len
#
# Authors
#       Ayose Cazorla <ayose.cazorla@hispalinux.es>
#
# This file is part of Pyslide.
#
# Pyslide is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Pyslide is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyslide; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                                
import pygame
import misc
import time
import Presentation
CS = Presentation.CurrentSurface

ALIGN_LEFT, ALIGN_CENTER, ALIGN_RIGHT = -1, 0, 1

class LastPoint:
    def __init__(self):
        self.reset()

    def reset(self):
        self.x, self.y = CS.map_units(20, 20)

    def set(self, x = None, y = None):
        if x is not None: self.x = x
        if y is not None: self.y = y

    def get(self):
        return self.x, self.y

class Item(pygame.sprite.Sprite):
    '''
    Item(attrs)

    This is the base class for all the items that you will put in
    the slide surface.
    '''

    def __init__ (self, attrs, last_point):
        pygame.sprite.Sprite.__init__(self)

        # default values
        self.open_effects = []
        self.close_effects = []
        self.current_effects = pygame.sprite.Group()
        self.closing = None
        self.ttl = None
        self.wait_time = None
        self.waiting = False
        self.raw_attrs = attrs
        self.keep_always = False
        self.align = ALIGN_LEFT
        self.list = None
        self.color = (255,255,255)

        self.orig_x, self.orig_y = last_point.get()
        self.parse_attrs(attrs)
        self.check_state()

        # move the last_point
        last_point.set(self.orig_x, self.orig_y + self.height)

    def parse_attrs(self, attrs):
        '''parse_attrs(attrs) -> None

        Loads the values of the attrs dictionary, and put the
        values as object attributes. This will convert everything
        (effects lists, integer values, etc). Unknown attributes
        will be  ignored.
        '''

        import Effects

        for r_key, val in attrs.items():
            key = r_key.lower()
            key_used = True

            if key == 'xy':
                try:
                    x, y = [int(x) for x in val.split(',', 1)]
                except ValueError:
                    raise misc.PyslideError, 'Invalid XY value: ' + str(val)
                else:
                    self.orig_x, self.orig_y = CS.map_units(x, y)

                    # save this positon.
                    global LAST_Y, LAST_X
                    LAST_X, LAST_Y = self.orig_x, self.orig_y
            elif key == 'y':
                if len(val) and val[0] in ('+', '-'):
                    relative = val[0] == '+' and 1 or -1
                    val = val[1:]
                else:
                    relative = 0

                try:
                    y = CS.map_units(y = int(val))
                except ValueError:
                    raise misc.PyslideError, 'Invalid y value: ' + str(val)

                if relative == -1:
                    self.orig_y -= y
                elif relative == 1:
                    self.orig_y += y
                else:
                    self.orig_y = y

            elif key == 'x':
                if len(val) and val[0] in ('+', '-'):
                    relative = val[0] == '+' and 1 or -1
                    val = val[1:]
                else:
                    relative = 0

                try:
                    x = CS.map_units(x = int(val))
                except ValueError:
                    raise misc.PyslideError, 'Invalid x value: ' + str(val)

                if relative == -1:
                    self.orig_x -= x
                elif relative == 1:
                    self.orig_x += x
                else:
                    self.orig_x = x

            elif key == 'color':
                c = misc.parse_color(val)
                if c is not None:
                    self.color = c
            elif key == 'ttl':
                if val == 'none':
                    self.ttl = None
                else:
                    try:
                        self.ttl = int(val) / 10.
                    except ValueError:
                        raise misc.PyslideError, 'You have to put an integer in TTL'
            elif key == 'wait':
                if val == 'none':
                    self.wait_time = None
                else:
                    try:
                        self.wait_time = int(val) / 10.
                    except ValueError:
                        raise misc.PyslideError, 'You have to put an integer in WAIT'
            elif key.startswith('open-effect-'):
                try:
                    ef = Effects.load(True, self, key[12:], val)
                except Effects.EffectError, e:
                    raise misc.PyslideError, 'Error in effect (%s): %s' % (key[12:], e)
                else:
                    self.open_effects.append(ef)
            elif key.startswith('close-effect-'):
                try:
                    ef = Effects.load(False, self, key[13:], val)
                except Effects.EffectError, e:
                    raise misc.PyslideError, 'Error in effect (%s): %s' % (key[13:], e)
                else:
                    self.close_effects.append(ef)
            elif key == 'keep-always':
                self.keep_always = val.lower() == 'yes'
            elif key == 'align':
                val = val.lower()
                if val == 'left':
                    self.align = ALIGN_LEFT
                elif val == 'center':
                    self.align = ALIGN_CENTER
                elif val == 'right':
                    self.align = ALIGN_RIGHT
                else:
                    raise misc.PyslideError, 'Invalid align value: ' + val
            elif key == 'list':
                val = val.lower()
                if val in ('circle', 'rhombus', 'square'):
                    self.list = val
                else:
                    raise misc.PyslideError, 'Invalid list type: ' + val
            else:
                key_used = False

            if key_used:
                del attrs[r_key]

    def check_state(self):
        '''check_state()

        This function is called after parse_attrs() to initialize
        the state of the item
        '''
        if self.align != ALIGN_LEFT:
            self.orig_x = 0

        image, rect = self.make_item()
        self.width, self.height = rect.size

        # is it a list?
        if self.list is not None:
            image, rect = self.make_list(image, rect)

        # align
        self.orig_x = self.align_x(rect.width)

    def make_item(self):
        '''make_item() -> (surface, rect)

        This function is called after a start() event. The child
        class will put in this method every initialization that 
        they need for the item. For instance, an image item will
        load its image in this method.

        It returns a tuple (surface, rect).
        '''
        return None, None

    def free_item(self):
        '''free_item(self) -> None

        The data that you set in make_item() have to be freed
        in this method.
        '''
        del self.image, self.rect

    def start(self):
        '''start() -> None

        start() is called every time that the item is put in a
        surface. It sets the required values for the item.
        '''

        self.closing = False
        if self.wait_time is None:
            self.start_effects()
        else:
            self.waiting = True
            self.start_time = time.time()
            self.hide()

    def start_effects(self):
        '''start_effects()

        Start all the open-effects
        '''

        self.waiting = False
        self.start_time = time.time()
        self.image, self.rect = self.make_item()

        # is it a list?
        if self.list is not None:
            self.image, self.rect = self.make_list(self.image, self.rect)

        if None in (self.image, self.rect):
            import warnings
            warnings.warn("Items can't return None in make_item", RuntimeWarning)

        for s in self.open_effects:
            self.current_effects.add(s)
            s.start()

    def align_x(self, width):
        '''align_x(width) -> x

        Moves the orig_x to aling the item
        '''
        max_width = pygame.display.get_surface().get_size()[0]

        if self.align == ALIGN_CENTER:
            x  = (max_width - width) / 2
        elif self.align == ALIGN_RIGHT:
            x = max_width - width
        elif self.align == ALIGN_LEFT:
            x = self.orig_x

        return x

    def make_list(self, image, rect):
        x, y = CS.map_units(10, 10)
        rect = pygame.Rect(rect) # copy it
        rect.width += 2*x + 10

        new_image = pygame.Surface(rect.size).convert_alpha()
        new_image.fill(0)
        new_image.blit(image, (2*x+10,0))

        if self.list == 'circle':
            pygame.draw.circle(new_image, self.color, (x, rect.height/2), y)
        elif self.list == 'rhombus':
            pygame.draw.polygon(new_image, self.color, [
                (x, rect.height/2 - y), 
                (0, rect.height/2),
                (x, rect.height/2 + y),
                (x*2, rect.height/2)])
        elif self.list in ('square', 'rect'):
            new_image.fill(self.color, (x/2, (rect.height-y)/2, x, y))
        else:
            raise NotImplementedError

        return new_image, rect

    def stop(self, show_effects=True):
        '''stop(show_effects=True) -> None
        
        stop() is called when the item is removed from the 
        surface. The item is not removed immediately, but it 
        loads the close_effects, and, when all are done, it
        kills itself.
        '''

        if self.keep_always:
            # do nothing is this in is a keep-always item
            return

        self.closing = True

        if self.waiting:
            self.kill()
            return

        to_close = self.close_effects[:]

        for s in self.current_effects.sprites():
            if s in self.open_effects:
                # if there is any open-effect, we don't
                # start the close-effects
                show_effects = False
            elif s in to_close:
                del to_close[to_close.index(s)]
            else:
                s.stop()

        if show_effects:
            for s in to_close:
                self.current_effects.add(s)
                s.start()
        else:
            self.kill()

    def force_stop(self):
        '''force_stop()

        Stop all the effects and kill itself. This does not show
        any close-effect
        '''

        for s in self.current_effects.sprites():
            s.stop()

        if not self.keep_always:
            self.kill()

    def kill(self):
        pygame.sprite.Sprite.kill(self)
        self.hide()

    def hide(self):
        '''hide()

        Hide the item. That is, creates a new surface (1x1), 
        will full alpha.
        '''
        # hide the item
        self.image = pygame.Surface((0,0))
        self.image.fill(0)
        self.rect = pygame.Rect(0,0,1,1)

    def update(self, *k):
        '''update()

        The main method. It is called in every iteration of the 
        program. It updates all the effects, and check the TTL and 
        WAIT values.

        Save some rare exceptions, children never need to implement 
        this method.
        '''
        del k

        if self.waiting:
            if time.time() - self.start_time > self.wait_time:
                self.start_effects()
            else:
                return

        if not self.closing and self.ttl is not None:
            if time.time() - self.start_time > self.ttl:
                self.stop()
                return

        # update the effects
        self.current_effects.update()
        if self.closing and len(self.current_effects) == 0:
            self.kill()

class ItemText(Item):
    '''ItemText(text, attrs)

    This implements the Text item
    '''

    def __init__ (self, text, *k):
        self.text = text
        self.shadow = (127,127,127)
        self.font_size = CS.map_units(y=60)
        self.font_file = None
        self.font = None
        self.interspace = 0

        Item.__init__(self, *k)

    def parse_attrs(self, attrs):
        'See Item.parse_attrs for info.'

        Item.parse_attrs(self, attrs)

        for key, val in attrs.items():
            key = key.lower()
            if key == 'shadow':
                if val.lower() == 'none':
                    self.shadow = None
                else:
                    c = misc.parse_color(val)
                    if c is not None: self.shadow = c
            elif key == 'font-file':
                self.font_file = val
            elif key == 'font-size':
                try:
                    self.font_size = CS.map_units(y=int(val))
                except ValueError:
                    raise misc.PyslideError, 'You have to put and integer in FONT-SIZE'
            elif key == 'interspace':
                try:
                    interspace = int(val)
                except ValueError:
                    raise misc.PyslideError, 'Invalid interspace value: ' + val

                self.interspace = CS.map_units(y = interspace)

    def make_item(self):
        'See Item.make_image for info.'
        colors = [self.color]
        if self.shadow is not None: colors.append(self.shadow)

        max_width = pygame.display.get_surface().get_size()[0]

        if hasattr(self, 'real_orig_x'):
            max_width -= self.real_orig_x
        else:
            max_width -= self.orig_x

        try:
            font = pygame.font.Font (self.font_file, self.font_size)
        except (IOError, RuntimeError), e:
            raise misc.PyslideError, 'loading font (%s): %s' \
                %(self.font_file, str(e))

        texts = []

        current_line = None
        line = self.text.replace('\r', ' ')  # HTML style
        for word in line.split():
            if current_line is None:
                current_line = word
            else:
                if font.size(current_line + ' ' + word)[0] > max_width:
                    texts.append(current_line)
                    current_line = word
                else:
                    current_line += ' ' + word

        if current_line:
            texts.append(current_line)
                
        # size of the item
        width, height = 0, 0
        for line in texts:
            w, h = font.size(line)

            height += h
            if w > width:
                width = w

        if len(colors) != 1:
            width += 2
            height += len(texts) * (2 + self.interspace)

        # draw the items
        image = pygame.Surface((width, height)).convert_alpha()
        image.fill(0)
        y = 0
        for line in texts:
            t1 = font.render (line, 1, colors[0])

            if len(colors) == 1:
                t = t1

            else:
                t2 = font.render (line, 1, colors[1])

                rect = t1.get_rect()
                rect = rect.inflate(4,4).move(2,2)

                t = pygame.Surface(rect.size).convert_alpha()
                t.fill(0) # full alpha
                t.blit(t2, (2,2))
                t.blit(t1, (0,0))

            w, h = t.get_size()
            if self.align == ALIGN_LEFT:
                x = 0
            elif self.align == ALIGN_CENTER:
                x = (width - w) / 2
            elif self.align == ALIGN_RIGHT:
                x = width - w
            else:
                raise NotImplementedError

            image.blit(t, (x,y))
            y += h + self.interspace

        rect = pygame.Rect(self.orig_x, self.orig_y, width, height)
        return image, rect


class ItemImage(Item):
    '''ItemImage

    This implements the Image item
    '''

    def __init__(self, *k):
        self.src = None
        self.dx = None
        self.dy = None

        Item.__init__(self, *k)

    def parse_attrs(self, attrs):
        Item.parse_attrs(self, attrs)

        for key, val in attrs.items():
            key = key.lower()
            if key == 'src':
                self.src = val
            elif key == 'size':
                try:
                    x, y = [int(x) for x in val.split(',', 1)]
                except ValueError:
                    raise misc.PyslideError, 'Invalid size value: ' + str(val)
                else:
                    self.dx, self.dy = x, y

    def make_item(self):
        # load the image
        image = misc.load_image(self.src).convert()

        if None not in (self.dx, self.dy):
            if self.dx < 10 or self.dy <10:
                size = image.get_size()
                if self.dx < 10:
                    self.dx *= size[0]
                if self.dy < 10:
                    self.dy *= size[1]

            mx, my = CS.map_units(self.dx, self.dy)
            image = pygame.transform.scale(image, (mx, my))

        # get the rect, using the size of the image
        rect = image.get_rect().move(self.orig_x, self.orig_y)
        return image, rect


class ItemShape(Item):
    def __init__(self, *k):
        self.s_shape = None
        self.s_center = None
        self.s_points = None
        self.s_radius = None
        self.s_start = None
        self.s_width = None
        self.s_end = None

        Item.__init__(self, *k)

    def parse_attrs(self, attrs):
        Item.parse_attrs(self, attrs)

        for key, val in attrs.items():
            key = key.lower()

            if key == 'shape':
                val = val.lower()
                if val in ('line', 'circle', 'arc', 'rect', 'polygon'):
                    self.s_shape = val
                else:
                    raise misc.PyslideError, 'UNKNOWN SHAPE: ' + str(val)

            elif key in ('start', 'end', 'points'):
                setattr(self, 's_' + key, val)

            elif key == 'center':
                try:
                    x, y = [int(x) for x in val.split(',', 1)]
                except ValueError:
                    raise misc.PyslideError, 'Invalid center value: ' + str(val)
                else:
                    self.s_center = CS.map_units(x, y)

            elif key == 'radius':
                try:
                    self.s_radius = CS.map_units(int(val))
                except ValueError:
                    raise misc.PyslideError, \
                        'Ivalid radius value. It has to be an integer: ' + str(val)

            elif key == 'width':
                try:
                    w = int(val)
                except ValueError:
                    raise misc.PyslideError, 'Invalid width: ' + str(val)
                else:
                    if w != 0:
                        self.s_width = max(1, CS.map_units(w))


        # interpret some valus depending on the shape
        if self.s_shape == 'arc':
            try:
                self.s_start = float(self.s_start)
                self.s_end = float(self.s_end)
            except ValueError, e:
                raise misc.PyslideError, \
                    'Invalid value for start/end angle: ' + str(e)

        elif self.s_shape in ('line', 'rect'):
            try:
                self.s_start = CS.map_units(*[int(x) for x in self.s_start.split(',', 1)])
                self.s_end = CS.map_units(*[int(x) for x in self.s_end.split(',', 1)])
            except ValueError, e:
                raise PyslideError, 'Invalid point value. ' + str(e)
        elif self.s_shape == 'polygon':
            p = []
            for point in self.s_points.split(';'):
                try:
                    x, y = [int(x) for x in point.split(',', 1)]
                except ValueError:
                    raise misc.PyslideError, 'Invalid point value: ' + str(point)
                else:
                    p.append(CS.map_units(x, y))

            self.s_points = p


        # default width
        if self.s_width is None:
            if self.s_shape == 'line':
                self.s_width = 1
            else:
                self.s_width = 0

    def make_item(self):

        if self.color == (0,0,0):
            colorkey = (255,255,255)
        else:
            colorkey = (0,0,0)

        surface = pygame.Surface(pygame.display.get_surface().get_size())
        surface.set_colorkey(colorkey)
        surface.fill(colorkey)
        rect = None

        if self.s_shape == 'line':
            rect = pygame.draw.line(surface, self.color,
                self.s_start, self.s_end, self.s_width)

        elif self.s_shape == 'circle':
            rect = pygame.draw.circle(surface, self.color,
                    self.s_center, self.s_radius, self.s_width)
        elif self.s_shape == 'arc':
            pass
        elif self.s_shape == 'rect':
            left, top = self.s_start
            right, bottom = self.s_end

            if left > right:
                left, right = right, left
            if top > bottom:
                top, bottom = bottom, top

            rect = pygame.Rect(left, top, right - left, bottom - top)

            if self.s_width == 0:
                surface.fill(self.color, rect)
            else:
                rect = pygame.draw.rect(surface, self.color, rect, 
                        self.s_width)
                
        elif self.s_shape == 'polygon':
            rect = pygame.draw.polygon(surface, self.color,
                    self.s_points, self.s_width)

        else:
            # hide the item
            image = pygame.Surface((0,0))
            image.fill(0)
            rect = pygame.Rect(0,0,1,1)

        if rect is None:
            rect = pygame.Rect(0,0,1,1)

        # copy the image. I did this in order to save memory
        image = pygame.Surface(rect.size)
        image.set_colorkey(colorkey)
        image.blit(surface.subsurface(rect), (0,0))

        self.orig_x, self.orig_y = rect.topleft

        return image, rect

