/*
 * 
 * Copyright (C) 2003 Ayose Cazorla Len
 * 
 * Authors
 *       Ayose Cazorla <ayose.cazorla@hispalinux.es>
 * 
 * This file is part of Pyslide.
 * 
 * Pyslide is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Pyslide is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Pyslide; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define NO_PYGAME_C_API
#include <pygame/pygame.h>

static PyObject* Pygame_SurfaceType;

/* set_alpha
 * 
 * This function was based on the pygame.surfarraw.pixels_alpha
 */
static char set_alpha_doc[] = 
    "set_alpha(surface, new_alpha) -> None\n\n"
    "Change the alpha of the surface.\nThis changes all the pixels\n\n"
    "This is something like\n\n"
    "   for pixel in pixels:\n"
    "       pixel.alpha *= new_alpha\n";

static PyObject *
set_alpha(PyObject *self, PyObject *args)
{
    SDL_Surface* surface;
    PyObject *surfobj;
    Uint8 *pixels;
    int len, x, offset;
    double alpha_multiply;

    if(!PyArg_ParseTuple(args, "O!d", Pygame_SurfaceType, &surfobj, &alpha_multiply))
        return NULL;

    surface = PySurface_AsSurface(surfobj);

    if(surface->format->BytesPerPixel != 4)
        return RAISE(PyExc_ValueError, "unsupport bit depth for alpha array");

    SDL_LockSurface(surface);

    if(surface->format->Amask == 0xff<<24)
        offset = SDL_BYTEORDER == SDL_LIL_ENDIAN ? 3 : 0;
    else if(surface->format->Amask == 0xff)
        offset = SDL_BYTEORDER == SDL_LIL_ENDIAN ? 0 : 3;
    else {
        SDL_UnlockSurface(surface);
        return RAISE(PyExc_ValueError, "unsupport colormasks for alpha reference array");
    }

    pixels = (Uint8*)surface->pixels;
    len = surface->w * surface->h * 4;
    for (x = 0; x < len; x += 4)
    {
        Uint8 a = pixels[x+offset];
        if (a == 0)
            pixels[x+offset] = 0;
        else
            pixels[x+offset] = (Uint8)(a * alpha_multiply);
    }

    SDL_UnlockSurface(surface);
    Py_INCREF(Py_None);
    return Py_None;
}

/*
 * set_valpha
 */
static char set_valpha_doc[] = "set_valpha(surface, new_alpha, limit) -> None\n";

static PyObject *
set_valpha(PyObject *self, PyObject *args)
{
    SDL_Surface* surface;
    PyObject *surfobj;
    Uint8 *pixels;
    int x, y, w, h, limit, direction, offset;
    double alpha;

    if(!PyArg_ParseTuple(args, "O!ii", 
                Pygame_SurfaceType, &surfobj, &limit, &direction))
        return NULL;

    surface = PySurface_AsSurface(surfobj);

    if(surface->format->BytesPerPixel != 4)
        return RAISE(PyExc_ValueError, "unsupport bit depth for alpha array");

    SDL_LockSurface(surface);

    if(surface->format->Amask == 0xff<<24)
        offset = SDL_BYTEORDER == SDL_LIL_ENDIAN ? 3 : 0;
    else if(surface->format->Amask == 0xff)
        offset = SDL_BYTEORDER == SDL_LIL_ENDIAN ? 0 : 3;
    else {
        SDL_UnlockSurface(surface);
        return RAISE(PyExc_ValueError, "unsupport colormasks for alpha reference array");
    }

    w = surface->w * 4;
    h = surface->h;
    if (direction > 0)
    {
        /* From top to bottom 
         */
        pixels = (Uint8*)surface->pixels + max(0, limit - 100)*w;
        for (y = max(0, limit - 100); y < h; y++)
        {
            if (y > limit)
            {
                for (x = 0; x < w; x+=4)
                    pixels[x+offset] = 0;
            } else
            {
                alpha = (limit - y) / 100.;
                for (x = 0; x < w; x+=4)
                    pixels[x+offset] = (Uint8)(pixels[x+offset] * alpha);
            }

            pixels += w;
        }
    } else
    {
        /* From bottom to up
         */
        pixels = (Uint8*)surface->pixels;
        h = min(limit + 100, h);
        for (y = 0; y < h; y++)
        {
            if (y > limit)
            {
                alpha = (y - limit) / 100.;
                for (x = 0; x < w; x+=4)
                    pixels[x+offset] = (Uint8)(pixels[x+offset] * alpha);
            } else
            {
                for (x = 0; x < w; x+=4)
                    pixels[x+offset] = 0;
            }

            pixels += w;
        }
    }

    SDL_UnlockSurface(surface);
    Py_INCREF(Py_None);
    return Py_None;
}

/*
 * set_halpha
 */
static char set_halpha_doc[] = "set_halpha(surface, limit, direction) -> None\n";

static PyObject *
set_halpha(PyObject *self, PyObject *args)
{
    SDL_Surface* surface;
    PyObject *surfobj;
    Uint8 *pixels;
    int x, y, w, h, limit, start, direction, offset;

    if(!PyArg_ParseTuple(args, "O!ii", 
                Pygame_SurfaceType, &surfobj, &limit, &direction))
        return NULL;

    surface = PySurface_AsSurface(surfobj);

    if(surface->format->BytesPerPixel != 4)
        return RAISE(PyExc_ValueError, "unsupport bit depth for alpha array");

    SDL_LockSurface(surface);

    if(surface->format->Amask == 0xff<<24)
        offset = SDL_BYTEORDER == SDL_LIL_ENDIAN ? 3 : 0;
    else if(surface->format->Amask == 0xff)
        offset = SDL_BYTEORDER == SDL_LIL_ENDIAN ? 0 : 3;
    else {
        SDL_UnlockSurface(surface);
        return RAISE(PyExc_ValueError, "unsupport colormasks for alpha reference array");
    }

    pixels = (Uint8*)surface->pixels;
    w = surface->w * 4;
    h = surface->h;
    limit *= 4;
    start = max(0, limit - 100);
    for (y = 0; y < h; y++)
    {
        if (direction > 0)
        {
            if (start < w && limit < w)
            {
                for(x = start; x < limit; x+=4)
                    pixels[x+offset] = (Uint8)(pixels[x+offset] * ((limit - x) / 100.));
                for(; x < w; x+=4)
                    pixels[x+offset] = 0;
            }
                
        } else
        {
            if (start > 0)
            {
                for (x = 0; x < min(w, start); x+=4)
                    pixels[x+offset] = 0;
                for (; x < min(w, start + 100); x+=4)
                    pixels[x+offset] = (Uint8)(pixels[x+offset] * ((x - start) / 100.));
            }
        }

        pixels += w;
    }

    SDL_UnlockSurface(surface);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyMethodDef cmisc_methods[] = {
    {"set_alpha",  set_alpha, METH_VARARGS, set_alpha_doc},
    {"set_valpha",  set_valpha, METH_VARARGS, set_valpha_doc},
    {"set_halpha",  set_halpha, METH_VARARGS, set_halpha_doc},
    {NULL, NULL, 0, NULL}        /* Sentinel */
};

void initcmisc(void)
{
    PyObject *pygame_surface;
    PyObject *dict;

    /* I need the pygame.surface.Surface type
     */

    if((pygame_surface = PyImport_ImportModule("pygame.surface")) == NULL)
        return;

    dict = PyModule_GetDict(pygame_surface);
    if((Pygame_SurfaceType = PyDict_GetItemString(dict, "Surface")) == NULL)
    {
        RAISE(PyExc_AttributeError, "pygame.surface module has no attribute 'Surface'");
        return;
    }

    Py_INCREF(Pygame_SurfaceType);
    Py_DECREF(pygame_surface);

    Py_InitModule("cmisc", cmisc_methods);
}

